#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

const char *name = "supermount-release";
const char *ctlfile = "/proc/fs/supermount/subfs";

int main (int argc, char *argv[])
{
	char buf[4200];
	char *p, *q, *r, *p2 = NULL, *q2 = NULL;
	int fd;
	struct stat st;
	unsigned int maj, min;
	unsigned int maj2, min2;
	uid_t uid;
	int i;

	/* boring initialization stuff */
	if (argc != 2) {
		fprintf(stderr, "Usage: %s devicename\n", name);
		exit(1);
	}

	/* become fully root */
	uid = getuid();
	i = setreuid(0, 0);
	if (i < 0) {
		fprintf(stderr, "%s: setreuid failed: %s\n", name, strerror(errno));
		exit(1);
	}

	/* read /proc/fs/supermount/subfs, terminate with a zero byte */
	fd = open(ctlfile, O_RDONLY);
	if (fd == 0 || fd == 1 || fd == 2) exit(1);
	if (fd < 0) {
		fprintf(stderr, "%s: cannot read %s\n", name, ctlfile);
		exit(1);
	}
	i = read(fd, buf, 4096);
	if (i < 0) {
		fprintf(stderr, "%s: error while reading from %s: %s\n", name, ctlfile, strerror(errno));
		exit(1);
	}
	if (i == 4096) {
		fprintf(stderr, "%s: %s is too long\n", name, ctlfile);
		exit(1);
	}
	if (memchr(buf, 0, i) != NULL) {
		fprintf(stderr, "%s: %s: faulty content\n", name, ctlfile);
		exit(1);
	}
	buf[i] = 0;
	close(fd);

	/* revert effective user id to the user */
	setreuid(0, uid);

	/* check the device argument */
	i = stat(argv[1], &st);
	if (i < 0) {
		fprintf(stderr, "%s: cannot stat %s: %s\n", name, argv[1], strerror(errno));
		exit(1);
	}
	if (!S_ISBLK(st.st_mode)) {
		fprintf(stderr, "%s: %s is not a block device\n", name, argv[1]);
		exit(1);
	}
	maj = major(st.st_rdev);
	min = minor(st.st_rdev);
	/* density variants of the same floppy drive should be treated the same */
	if (maj == 2) min &= ~124;

	/* parse the buffer holding the contents of /proc/fs/supermount/subfs */
	p = buf;
	while (*p != 0) {
		q = strchr(p, ' ');
		r = strchr(p, '\n');
		if (*p != '/' || q == NULL || r == NULL || q > r) {
			fprintf(stderr, "%s: %s: faulty content\n", name, ctlfile);
			exit(1);
		}
		*q = 0;
		i = stat(p, &st);
		if (i < 0) {
			p = r + 1;
			continue;
		}
		if (!S_ISBLK(st.st_mode)) {
			p = r + 1;
			continue;
		}
		maj2 = major(st.st_rdev);
		min2 = minor(st.st_rdev);
		if (maj2 == 2) min2 &= ~124;
		if (maj2 != maj || min2 != min) {
			p = r + 1;
			continue;
		}
		if (p2 != NULL) {
			fprintf(stderr, "%s: more entries matching %s found in supermount list\n", name, argv[0]);
			exit(1);
		}
		p2 = p;
		q2 = q;
		p = r + 1;
	}
	if (p2 == NULL) {
		fprintf(stderr, "%s: %s not found in supermount list\n", name, argv[0]);
		exit(1);
	}

	/* prepare the command */
	strcpy(q2, " release force\n");

	/* become fully root again */
	i = setreuid(0, 0);
	if (i < 0) {
		fprintf(stderr, "%s: setreuid failed: %s\n", name, strerror(errno));
		exit(1);
	}

	/* execute the command */
	fd = open(ctlfile, O_WRONLY);
	if (fd == 0 || fd == 1 || fd == 2) exit(1);
	if (fd < 0) {
		fprintf(stderr, "%s: cannot write %s\n", name, ctlfile);
		exit(1);
	}
	i = write(fd, p2, strlen(p2));
	if (i < 0) {
		fprintf(stderr, "%s: error while writing to %s: %s\n", name, ctlfile, strerror(errno));
		exit(1);
	}
	if (i != strlen(p2)) {
		fprintf(stderr, "%s: short write to %s\n", name, ctlfile);
		exit(1);
	}
	close(fd);
	return 0;
}
